/** 
 * @file  RegistryBundle.h
 *
 * @brief Declaration of CRegKeyEx C++ wrapper class for reading Windows registry
 */
// RCS ID line follows -- this is updated by CVS
// $Id: RegistryBundle.h,v 1.4 2006/03/04 17:54:58 elsapo Exp $

#pragma once

#include <winreg.h>

#define REINTERPRET_CONVERT(type,data) (*((type*)&data))

class RegVal
{
  DWORD m_dwAllocSize;
  DWORD dwType;
  ULARGE_INTEGER val;
  DWORD dwSize;
  DWORD AllocateMemory(DWORD size = 0, bool bRealloc = false);
public:
  DWORD DataType() const { return dwType; }
  DWORD DataSize() const { return dwSize; }
  DWORD StringDataLength() const { return (IsString()) ? dwSize/sizeof(TCHAR) : 0; }
  bool Clear();
  RegVal(DWORD type = REG_NONE, DWORD size = 0) : m_dwAllocSize(0), dwType(REG_NONE), dwSize(0)
  {
    val.QuadPart = 0;
    if (type != REG_NONE)
      Initialize(type, size);
  }
  ~RegVal()
  {
    Clear();
  }
  RegVal(const RegVal &src);
  bool Initialize(DWORD type = REG_NONE, DWORD size = 0);
  bool IsNumeric() const
  {
    return (dwType == REG_DWORD_LITTLE_ENDIAN) || (dwType == REG_DWORD_BIG_ENDIAN) || (dwType == REG_QWORD_LITTLE_ENDIAN);
  }
  bool IsString() const
  {
    return (dwType == REG_SZ) || (dwType == REG_EXPAND_SZ) || (dwType == REG_MULTI_SZ);
  }
  bool IsStringSingle() const
  {
    return (dwType == REG_SZ) || (dwType == REG_EXPAND_SZ);
  }
  bool IsStringMulti() const
  {
    return (dwType == REG_MULTI_SZ);
  }
  bool IsBinary() const
  {
    return dwType == REG_BINARY;
  }
  DWORD DwordValue() const { return ((dwType == REG_DWORD_LITTLE_ENDIAN) || (dwType == REG_DWORD_BIG_ENDIAN)) ? val.LowPart : 0; }
  ULONGLONG QwordValue() const { return (dwType == REG_QWORD_LITTLE_ENDIAN) ? val.QuadPart : 0; }
  float FloatValue() const { return (dwType == REG_DWORD) ? REINTERPRET_CONVERT(float, val.LowPart) : 0.0f; }
  double DoubleValue() const { return (dwType == REG_QWORD) ? REINTERPRET_CONVERT(double, val.QuadPart) : 0.0; }
  int IntValue() const { return (dwType == REG_DWORD) ? REINTERPRET_CONVERT(int, val.LowPart) : 0; }
  LONGLONG LongValue() const { return (dwType == REG_QWORD) ? REINTERPRET_CONVERT(LONGLONG, val.QuadPart) : 0; }
  bool BoolValue() const { return DwordValue(); }
  DWORD StringValue(LPTSTR lpBuffer, DWORD dwBufferCharLength) const;
  DWORD BinaryValue(LPBYTE lpBuffer, DWORD dwBufferSize) const;
  DWORD StringValue(LPTSTR &lpBuffer) const;
  DWORD BinaryValue(LPBYTE &lpBuffer) const;
  const BYTE* RawValue() const
  {
    return (IsNumeric()) ? reinterpret_cast<const BYTE*>(&val.LowPart) : reinterpret_cast<const BYTE*>(val.LowPart);
  }
  operator DWORD() const { return DwordValue(); }
  operator ULONGLONG() const { return QwordValue(); }
  operator float() const { return FloatValue(); }
  operator double() const { return DoubleValue(); }
  operator int() const { return IntValue(); }
  operator LONGLONG() const { return LongValue(); }
  operator bool() const { return BoolValue(); }
  void DwordValue(DWORD dwValue, bool bBigEndian = false)
  {
    Clear();
    dwType = (bBigEndian) ? REG_DWORD_BIG_ENDIAN : REG_DWORD;
    val.LowPart = dwValue;
    dwSize = sizeof(dwValue);
  }
  void QwordValue(ULONGLONG QuadValue)
  {
    Clear();
    dwType = REG_QWORD;
    val.QuadPart = QuadValue;
    dwSize = sizeof(QuadValue);
  }
  void FloatValue(float fValue)
  {
    DwordValue(REINTERPRET_CONVERT(DWORD, fValue));
  }
  void DoubleValue(double dValue)
  {
    QwordValue(REINTERPRET_CONVERT(ULONGLONG, dValue));
  }
  void IntValue(int iValue)
  {
    DwordValue(REINTERPRET_CONVERT(DWORD, iValue));
  }
  void LongValue(LONGLONG lValue)
  {
    QwordValue(REINTERPRET_CONVERT(ULONGLONG, lValue));
  }
  void BoolValue(bool bValue)
  {
    DwordValue((DWORD)bValue);
  }
  bool StringValue(LPCTSTR lpBuffer, DWORD dwBufferCharLength, bool bRealloc, DWORD dwStringType = REG_SZ);
  bool BinaryValue(const BYTE *lpBuffer, DWORD dwBufferSize, bool bRealloc);
};

class CRegKeyEx
{
//protected:
	HKEY 	m_hKey;
  HKEY 	m_hRootKey;
	TCHAR m_sPath[_MAX_PATH];
  TCHAR m_sMachine[_MAX_PATH];
  LSTATUS m_OperationStatus;
  bool m_bOwned;
  LSTATUS OpenRootKey(HKEY hKeyRoot, LPCTSTR lpRemoteMachine = NULL);
public:
  CRegKeyEx() : m_hKey(NULL), m_hRootKey(NULL), m_bOwned(false), m_OperationStatus(ERROR_SUCCESS)
  {
    m_sPath[0] = 0;
    m_sMachine[0] = 0;
  }
	~CRegKeyEx()
  {
    Close();
  }
  LSTATUS OperationStatus() const { return m_OperationStatus; }
  LSTATUS WriteValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, DWORD dwType, const BYTE* lpData, DWORD cbData);
  LSTATUS WriteValue(LPCTSTR pszValueName, DWORD dwType, const BYTE* lpData, DWORD cbData)
  {
    return WriteValue(NULL, pszValueName, dwType, lpData, cbData);
  }
  LSTATUS ReadValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPDWORD lpdwType, LPBYTE lpData, DWORD &cbData) const;
  LSTATUS ReadValue(LPCTSTR pszValueName, LPDWORD lpdwType, LPBYTE lpData, DWORD &cbData) const
  {
    return ReadValue(NULL, pszValueName, lpdwType, lpData, cbData);
  }
  LSTATUS WriteValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, const RegVal *rvData);
  LSTATUS WriteValue(LPCTSTR pszValueName, const RegVal *rvData)
  {
    return WriteValue(NULL, pszValueName, rvData);
  }
  LSTATUS ReadValue(LPCTSTR pszSubKey, LPCTSTR pszValueName, RegVal *rvData) const;
  LSTATUS ReadValue(LPCTSTR pszValueName, RegVal *rvData) const
  {
    return ReadValue(NULL, pszValueName, rvData);
  }
// Operations
	HKEY GetKey() const { return m_hKey; }
  operator HKEY() const { return m_hKey; }
  LPCTSTR GetKeyPath() const { return m_sPath; }
  operator LPCTSTR() const { return m_sPath; }
  LPCTSTR GetMachineName() const { return m_sMachine; }
	void Close(bool bOnlyCurrent = false);
	LSTATUS Open(HKEY hKeyRoot, LPCTSTR pszPath, LPCTSTR lpRemoteMachine = NULL);
	LSTATUS OpenWithAccess(HKEY hKeyRoot, LPCTSTR pszPath, REGSAM regsam = KEY_QUERY_VALUE, LPCTSTR lpRemoteMachine = NULL);
	LSTATUS OpenNoCreateWithAccess(HKEY hKeyRoot, LPCTSTR pszPath, REGSAM regsam = KEY_QUERY_VALUE, LPCTSTR lpRemoteMachine = NULL);
	bool OpenRegMachine(LPCTSTR key, REGSAM regsam = KEY_QUERY_VALUE, LPCTSTR lpRemoteMachine = NULL, bool bCreate = true);
	bool OpenRegUser(LPCTSTR key, REGSAM regsam = KEY_QUERY_VALUE, LPCTSTR lpRemoteMachine = NULL, bool bCreate = true);
  bool QueryRegMachine(LPCTSTR key, REGSAM regsam = KEY_QUERY_VALUE, LPCTSTR lpRemoteMachine = NULL)
  {
    return OpenRegMachine(key, regsam, lpRemoteMachine, false);
  }
	bool QueryRegUser(LPCTSTR key, REGSAM regsam = KEY_QUERY_VALUE, LPCTSTR lpRemoteMachine = NULL)
  {
    return OpenRegUser(key, regsam, lpRemoteMachine, false);
  }

  LSTATUS WriteBinary(LPCTSTR pszSubKey, LPCTSTR pszValueName, const BYTE* lpData, DWORD cbData)
  {
    return WriteValue(pszSubKey, pszValueName, REG_BINARY, lpData, cbData);
  }
  LSTATUS WriteBinary(LPCTSTR pszValueName, const BYTE* lpData, DWORD cbData)
  {
    return WriteBinary(NULL, pszValueName, lpData, cbData);
  }

	LSTATUS WriteDword(LPCTSTR pszSubKey, LPCTSTR pszValueName, DWORD dwVal)
  {
    return WriteValue(pszSubKey, pszValueName, REG_DWORD, (const BYTE*) &dwVal, sizeof(dwVal));
  }
  LSTATUS WriteDword(LPCTSTR pszValueName, DWORD dwVal)
  {
    return WriteDword(NULL, pszValueName, dwVal);
  }

  LSTATUS WriteQword(LPCTSTR pszSubKey, LPCTSTR pszValueName, ULONGLONG qwVal)
  {
    return WriteValue(pszSubKey, pszValueName, REG_QWORD, (const BYTE*) &qwVal, sizeof(qwVal));
  }
  LSTATUS WriteQword(LPCTSTR pszValueName, ULONGLONG qwVal)
  {
    return WriteQword(NULL, pszValueName, qwVal);
  }

	LSTATUS WriteBool(LPCTSTR pszSubKey, LPCTSTR pszValueName, bool bVal)
  {
    return WriteDword(pszSubKey, pszValueName, (DWORD)bVal);
  }
  LSTATUS WriteBool(LPCTSTR pszValueName, bool bVal)
  {
    return WriteBool(NULL, pszValueName, bVal);
  }

  LSTATUS WriteInt(LPCTSTR pszSubKey, LPCTSTR pszValueName, int iVal)
  {
    return WriteDword(pszSubKey, pszValueName, REINTERPRET_CONVERT(DWORD, iVal));
  }
  LSTATUS WriteInt(LPCTSTR pszValueName, int iVal)
  {
    return WriteInt(NULL, pszValueName, iVal);
  }

  LSTATUS WriteLong(LPCTSTR pszSubKey, LPCTSTR pszValueName, LONGLONG iVal)
  {
    return WriteQword(pszSubKey, pszValueName, REINTERPRET_CONVERT(ULONGLONG, iVal));
  }
  LSTATUS WriteLong(LPCTSTR pszValueName, LONGLONG iVal)
  {
    return WriteLong(NULL, pszValueName, iVal);
  }

	LSTATUS WriteFloat(LPCTSTR pszSubKey, LPCTSTR pszValueName, float fVal)
  {
    return WriteDword(pszSubKey, pszValueName, REINTERPRET_CONVERT(DWORD, fVal));
  }
  LSTATUS WriteFloat(LPCTSTR pszValueName, float fVal)
  {
    return WriteFloat(NULL, pszValueName, fVal);
  }

  LSTATUS WriteDouble(LPCTSTR pszSubKey, LPCTSTR pszValueName, double dVal)
  {
    return WriteQword(pszSubKey, pszValueName, REINTERPRET_CONVERT(ULONGLONG, dVal));
  }
  LSTATUS WriteDouble(LPCTSTR pszValueName, double dVal)
  {
    return WriteDouble(NULL, pszValueName, dVal);
  }

	LSTATUS WriteString(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPCTSTR pszVal);
  LSTATUS WriteString(LPCTSTR pszValueName, LPCTSTR pszVal)
  {
    return WriteString(NULL, pszValueName, pszVal);
  }

  LSTATUS WriteMultiString(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPCTSTR pszVal, DWORD dwCharLen);
  LSTATUS WriteMultiString(LPCTSTR pszValueName, LPCTSTR pszVal, DWORD dwCharLen)
  {
    return WriteMultiString(NULL, pszValueName, pszVal, dwCharLen);
  }

  template<class T>
  T ReadNumeric(LPCTSTR pszSubKey, LPCTSTR pszValueName, T defval = (T)0, LPDWORD lpdwType = NULL) const
  {
	  //assert(m_hKey);
	  //assert(pszKey);

	  T dwDest;
	  DWORD dwSize = sizeof(dwDest);

	  LSTATUS operationStatus = ReadValue(pszSubKey, pszValueName, lpdwType, (LPBYTE) &dwDest, dwSize);

	  if (operationStatus == ERROR_SUCCESS)
		  return dwDest;
	  return defval;
  }
  template<class T>
  T ReadNumeric(LPCTSTR pszValueName, T defval = (T)0, LPDWORD lpdwType = NULL) const
  {
    return ReadNumeric(NULL, pszValueName, defval, lpdwType);
  }
	/*DWORD ReadDword (LPCTSTR pszKey, DWORD defval);
	float ReadFloat (LPCTSTR pszKey, float defval);
	BOOL ReadBool(LPCTSTR pszKey, BOOL defval);
	LONG ReadLong (LPCTSTR pszKey, LONG defval);
	UINT ReadUint (LPCTSTR pszKey, UINT defval);
	int ReadInt (LPCTSTR pszKey, int defval);
	short int ReadShort (LPCTSTR pszKey, short int defval);
	BYTE ReadByte (LPCTSTR pszKey, BYTE defval);*/
	CString ReadString(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPCTSTR defval) const;
  CString ReadString(LPCTSTR pszValueName = NULL, LPCTSTR defval = NULL) const
  {
    return ReadString(NULL, pszValueName, defval);
  }
	DWORD ReadChars(LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pData = NULL, DWORD dwCharLength = 0) const;
  DWORD ReadChars(LPCTSTR pszValueName, LPTSTR pData = NULL, DWORD dwCharLength = 0) const
  {
    return ReadChars(NULL, pszValueName, pData, dwCharLength);
  }
  LSTATUS EnumValue(DWORD dwIndex,LPTSTR lpValueName,LPDWORD lpcValueName,LPDWORD lpType = NULL,LPBYTE lpData = NULL,LPDWORD lpcbData = NULL) const;
  LSTATUS EnumValueFirst(LPTSTR lpValueName,LPDWORD lpcValueName,LPDWORD lpType = NULL,LPBYTE lpData = NULL,LPDWORD lpcbData = NULL) const
  {
    return EnumValue(0, lpValueName, lpcValueName, lpType, lpData, lpcbData);
  }
  LSTATUS EnumValue(DWORD dwIndex,LPTSTR lpValueName,LPDWORD lpcValueName, RegVal *rvData) const;
  LSTATUS EnumValueFirst(LPTSTR lpValueName,LPDWORD lpcValueName, RegVal *rvData) const
  {
    return EnumValue(0, lpValueName, lpcValueName, rvData);
  }
  LSTATUS DeleteKey();
  LSTATUS DeleteTree();
  LSTATUS DeleteValue(LPCTSTR pszValueName = NULL);
  LSTATUS DeleteKeyValue(LPCTSTR pszSubKey, LPCTSTR pszValueName = NULL);
  LSTATUS QueryInfoKey(LPTSTR lpClass = NULL,LPDWORD lpcClass = NULL,LPDWORD lpcSubKeys = NULL,LPDWORD lpcMaxSubKeyLen = NULL,LPDWORD lpcMaxClassLen = NULL,LPDWORD lpcValues = NULL,LPDWORD lpcMaxValueNameLen = NULL,LPDWORD lpcMaxValueLen = NULL,LPDWORD lpcbSecurityDescriptor = NULL,PFILETIME lpftLastWriteTime = NULL) const;
};
