/** 
 * @file  CompareOptions.cpp
 *
 * @brief Compare options implementation.
 */
// ID line follows -- this is updated by SVN
// $Id: ByteComparator.h 3397 2006-07-27 10:41:24Z kimmov $

#include "stdafx.h"
#include "Diff.h"
#include "CompareOptions.h"

// Global defined in diffutils code
//extern int recursive;

/**
 * @brief Default constructor.
 */
CompareOptions::CompareOptions()
: m_ignoreWhitespace(WHITESPACE_COMPARE_ALL)
, m_bIgnoreBlankLines(false)
, m_bIgnoreCase(false)
, m_bIgnoreEOLDifference(false)
{
}

/**
 * @brief Copy constructor.
 */
CompareOptions::CompareOptions(const CompareOptions & options)
: m_ignoreWhitespace(options.m_ignoreWhitespace)
, m_bIgnoreBlankLines(options.m_bIgnoreBlankLines)
, m_bIgnoreCase(options.m_bIgnoreCase)
, m_bIgnoreEOLDifference(options.m_bIgnoreEOLDifference)
{
}

/**
 * @brief Sets options from DIFFOPTIONS structure.
 * @param [in] options Diffutils options.
 */
void CompareOptions::SetFromDiffOptions(const DIFFOPTIONS &options)
{
	switch (options.nIgnoreWhitespace)
	{
	case 0:
		m_ignoreWhitespace = WHITESPACE_COMPARE_ALL;
		break;
	case 1:
		m_ignoreWhitespace = WHITESPACE_IGNORE_CHANGE;
		break;
	case 2:
		m_ignoreWhitespace = WHITESPACE_IGNORE_ALL;
		break;
	default:
		_RPTF0(_CRT_ERROR, "Unknown whitespace ignore value!");
		break;
	}
	m_bIgnoreBlankLines = options.bIgnoreBlankLines;
	m_bIgnoreCase = options.bIgnoreCase;
	m_bIgnoreEOLDifference = options.bIgnoreEol;
}

/**
 * @brief Default constructor.
 */
QuickCompareOptions::QuickCompareOptions()
: m_bStopAfterFirstDiff(FALSE)
{

}

/**
 * @brief Default constructor.
 */
DiffutilsOptions::DiffutilsOptions()
: m_outputStyle(DIFF_OUTPUT_NORMAL)
, m_contextLines(0)
, m_filterCommentsLines(0)
{
}

/**
 * @brief Constructor cloning CompareOptions.
 * @param [in] options CompareOptions instance to clone.
 */
DiffutilsOptions::DiffutilsOptions(const CompareOptions& options)
: CompareOptions(options)
, m_outputStyle(DIFF_OUTPUT_NORMAL)
, m_contextLines(0)
, m_filterCommentsLines(0)
{
}

/**
 * @brief Sets options from DIFFOPTIONS structure.
 * @param [in] options Diffutils options.
 */
void DiffutilsOptions::SetFromDiffOptions(const DIFFOPTIONS & options)
{
	CompareOptions::SetFromDiffOptions(options);
	m_filterCommentsLines = options.bFilterCommentsLines;
}

/**
 * @brief Set options to diffutils.
 * Diffutils uses options from global variables we must set. Those variables
 * aren't most logically named, and most are just int's, with some magic
 * values and meanings, with fancy combinations? So not easy to setup. This
 * function maps our easier to handle compare options to diffutils globals.
 */
void DiffutilsOptions::SetToDiffUtils()
{
	/*switch (m_outputStyle)
	{
	case DIFF_OUTPUT_NORMAL:
		output_style = OUTPUT_NORMAL;
		break;
	case DIFF_OUTPUT_CONTEXT:
		output_style = OUTPUT_CONTEXT;
		break;
	case DIFF_OUTPUT_UNIFIED:
		output_style = OUTPUT_UNIFIED;
		break;
	default:
		_RPTF0(_CRT_ERROR, "Unknown output style!");
		break;
	}*/
  if ((m_outputStyle >= DIFF_OUTPUT_UNSPECIFIED) && (m_outputStyle <= DIFF_OUTPUT_UNIFIED))
    output_style = (enum output_style)m_outputStyle;
  else
    _RPTF0(_CRT_ERROR, "Unknown output style!");

	context = m_contextLines;
	if ((m_ignoreWhitespace != WHITESPACE_COMPARE_ALL) || m_bIgnoreEOLDifference)
    ignore_white_space = IGNORE_TAB_EXPANSION;
	if (m_ignoreWhitespace == WHITESPACE_IGNORE_CHANGE)
    ignore_white_space = IGNORE_SPACE_CHANGE;
	if (m_ignoreWhitespace == WHITESPACE_IGNORE_ALL)
    ignore_white_space = IGNORE_ALL_SPACE;
	ignore_blank_lines = m_bIgnoreBlankLines;
	ignore_case = m_bIgnoreCase;
	/*ignore_eol_diff*/ strip_trailing_cr = m_bIgnoreEOLDifference;
	//length_varies = (m_ignoreWhitespace != WHITESPACE_COMPARE_ALL);

	// We have no interest changing these values, hard-code them.
	/*always_text_flag*/ text = false; // diffutils needs to detect binary files
	horizon_lines = 0;
	/*heuristic*/ speed_large_files = false; //true;
	recursive = false;
}

/**
 * @brief Gets options to DIFFOPTIONS structure.
 * @param [out] options Diffutils options.
 */
void DiffutilsOptions::GetAsDiffOptions(DIFFOPTIONS &options)
{
	options.bFilterCommentsLines = m_filterCommentsLines;
	options.bIgnoreBlankLines = m_bIgnoreBlankLines;
	options.bIgnoreCase = m_bIgnoreCase;
	options.bIgnoreEol = m_bIgnoreEOLDifference;
	
	/*switch (m_ignoreWhitespace)
	{
	case WHITESPACE_COMPARE_ALL:
		options.nIgnoreWhitespace = 0;
		break;
	case WHITESPACE_IGNORE_CHANGE:
		options.nIgnoreWhitespace = 1;
		break;
	case WHITESPACE_IGNORE_ALL:
		options.nIgnoreWhitespace = 2;
		break;
	default:
		_RPTF0(_CRT_ERROR, "Unknown whitespace ignore value!");
		break;
	}*/
  if ((m_ignoreWhitespace >= WHITESPACE_COMPARE_ALL) && (m_ignoreWhitespace <= WHITESPACE_IGNORE_ALL))
    options.nIgnoreWhitespace = (int)m_ignoreWhitespace;
  else
    _RPTF0(_CRT_ERROR, "Unknown whitespace ignore value!");
}
