/** 
 * @file  stringdiffs.cpp
 *
 * @brief Implementation file for sd_ComputeWordDiffs (q.v.)
 *
 */
// RCS ID line follows -- this is updated by CVS
// $Id: stringdiffs.cpp 4348 2007-06-25 17:04:14Z kimmov $

#include "stdafx.h"
#include <mbctype.h>
#include "stringdiffs.h"
#include "CompareOptions.h"
#include "stringdiffsi.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

LPCTSTR constWordBreaks = _T(".,;:()[]{}!@#\"$%^&*~+-=<>\'/\\|");

/** Does character introduce a multicharacter character? */
#ifdef UNICODE
#define IsLeadByte(ch) false
#else
#define IsLeadByte(ch) (_getmbcp() && IsDBCSLeadByte(ch))
#endif

/**
 * @brief Is it whitespace (excludes all lead & trail bytes)?
 */
bool isSafeWhitespace(TCHAR ch)
{
	return xisspace(ch) && !IsLeadByte(ch);
}

/**
 * @brief Is it a non-whitespace wordbreak character (ie, punctuation)?
 */
bool isWordBreak(int breakType, TCHAR ch)
{
	// breakType==0 means whitespace only
	if (breakType == 0) return false;
	// breakType==1 means break also on punctuation
	return _tcschr(constWordBreaks, ch);
}

/**
 * @brief adjust the range of the specified word diffs down to byte level.
 * @param str1, str2 [in] line to be compared
 * @param casitive [in] true for case-sensitive, false for case-insensitive
 * @param xwhite [in] This governs whether we handle whitespace specially (see WHITESPACE_COMPARE_ALL, WHITESPACE_IGNORE_CHANGE, WHITESPACE_IGNORE_ALL)
 */
void wordLevelToByteLevel(wdiffarray * pDiffs, const CString& str1, const CString& str2, bool casitive, int xwhite)
{
  int cc3 = pDiffs->GetSize();
	for (int i = 0; i < cc3; ++i)
	{
		int begin1, begin2, end1, end2;
		wdiff *pDiff = &(*pDiffs)[i];
		CString str1_2, str2_2;
		str1_2 = str1.Mid(pDiff->start[0], pDiff->GetLength(0));
		str2_2 = str2.Mid(pDiff->start[1], pDiff->GetLength(1));
		sd_ComputeByteDiff(str1_2, str2_2, casitive, xwhite, begin1, begin2, end1, end2);
		if (begin1 < 0)
		{
			// no visible diff on side1
			pDiff->start[0] = str1.GetLength();
			pDiff->end[0] = pDiff->start[0] - 1;
		}
		else
		{
			pDiff->end[0] = pDiff->start[0] + end1;
			pDiff->start[0] += begin1;
		}
		if (begin2 < 0)
		{
			// no visible diff on side2
			pDiff->start[1] = str2.GetLength();
			pDiff->end[1] = pDiff->start[1] - 1;
		}
		else
		{
			pDiff->end[1] = pDiff->start[1] + end2;
			pDiff->start[1] += begin2;
		}
	}
}

/**
 * @brief Construct our worker object and tell it to do the work
 */
float
sd_ComputeWordDiffs(const CString & str1, const CString & str2, wdiffarray * pDiffs,
	bool case_sensitive, int whitespace, int breakType, bool byte_level,
	bool bComputeDiffStatus)
{
  float res = 1.0f;
	stringdiffs sdiffs(str1, str2, case_sensitive, whitespace, breakType, pDiffs);
	// Hash all words in both lines and then compare them word by word
	// storing differences into m_wdiffs
	sdiffs.BuildWordDiffList();
	// Adjust the range of the word diff down to byte level.
  if (bComputeDiffStatus)
  {
    CString str10 = str1;
    CString str20 = str2;
    int l1 = str10.Trim().GetLength();
    int l2 = str20.Trim().GetLength();
    if (l1 && l2)
    {
      int k = (int)(l2 > l1);
      int l = (k) ? l2 : l1;
      if (l)
      {
        int c = pDiffs->GetSize();
        int h = 0;
        for (int i = 0; i < c; ++i)
          h += pDiffs->GetAt(i).GetLength(k);
        res = h / (float)l;
      }
      ASSERT(res <= 1.0f);
    }
    else
      if (l1 == l2)
        res = 0.0f;
  }
	if (byte_level)
		wordLevelToByteLevel(pDiffs, str1, str2, case_sensitive, whitespace);
  return res;
}

/**
 * @brief stringdiffs constructor simply loads all members from arguments
 */
stringdiffs::stringdiffs(const CString & str1, const CString & str2,
	bool case_sensitive, int whitespace, int breakType,
	wdiffarray * pDiffs)
: m_str1(str1)
, m_str2(str2)
, m_case_sensitive(case_sensitive)
, m_whitespace(whitespace)
, m_breakType(breakType)
, m_pDiffs(pDiffs)
{
}

bool stringdiffs::IsRegularWord(int windex, int index) const
{
  if (windex < 0)
    windex = 0;
  const wordarray *wds;
  const CString *str;
  if (index)
  {
    wds = &m_words2;
    str = &m_str2;
  }
  else
  {
    wds = &m_words1;
    str = &m_str1;
  }
  int c = wds->GetSize();
  if (windex < c)
  {
    word wd = wds->GetAt(windex);
    return (wd.length() > 1) || !isWordBreak(m_breakType, str->GetAt(wd.start));
  }
  return false;
}

// @brief Find closest matching word
bool stringdiffs::findSync(int &w1, int &w2) const
{
  int w1c = w1;
  int w2c = w2;
	// Look among remaining words in m_words2 for a word that matches w1
	int cw2 = -1;
  int cw22 = -1;
  int cc1 = m_words1.GetSize();
  for (int i = w1; i < cc1; ++i)
    if (IsRegularWord(i, 0))
    {
      w1 = i;
      break;
    }
  int w10 = w1;
  int w1b = w1;
	while (w1 < cc1)
	{
    if (IsRegularWord(w1, 0))
      cw22 = FindNextMatchInWords2(m_words1[w1], w2);
    if (cw2 < 0)
    {
      cw2 = (cw22>=0) ? cw22 : FindNextMatchInWords2(m_words1[w1], w2);
      w1b = w1;
    }
		if (cw2>=0 && cw22>=0)
			break;
		// No word matches w1
		++w1;
	}
	// Look among remaining words in m_words1 for a word that matches w2
	int cw1 = -1;
  int cw12 = -1;
  cc1 = m_words2.GetSize();
  for (int i = w2; i < cc1; ++i)
    if (IsRegularWord(i, 1))
    {
      w2 = i;
      break;
    }
  int w2b = w2;
	while (w2 < cc1)
	{
    if (IsRegularWord(w2, 1))
      cw12 = FindNextMatchInWords1(m_words2[w2], w10);
		if (cw1 < 0)
    {
      cw1 = (cw12>=0) ? cw12 : FindNextMatchInWords1(m_words2[w2], w10);
      w2b = w2;
    }
		if (cw1>=0 && cw12>=0)
			break;
		// No word matches w2
		++w2;
	}
  if (cw22>=0)
  {
    while (cw22 > w2c && w1 > w1c && AreWordsSame(m_words1[w1-1], m_words2[cw22-1]))
    {
      --w1;
      --cw22;
    }
    cw2 = cw22;
  }
  else
    w1 = w1b;
  if (cw12>=0)
  {
    while (cw12 > w1c && w2 > w2c && AreWordsSame(m_words1[cw12-1], m_words2[w2-1]))
    {
      --cw12;
      --w2;
    }
    cw1 = cw12;
  }
  else
    w2 = w2b;
	if (cw1 == -1)
	{
		if (cw2 == -1)
			return false;
		w2 = cw2;
	}
	else
	{
		if (cw2 == -1)
			w1 = cw1;
		else
		{
			// We have candidates advancing along either string
			// Pick closer
			int len1 = m_words1[cw1].end - m_words1[w1].start;
			int len2 = m_words2[cw2].end - m_words2[w2].start;
			if (len1 < len2)
				w1 = cw1;
			else
				w2 = cw2;
		}
	}
	return true;
}

// @brief Find next word in m_words2 (starting at bw2) that matches needword1 (in m_words1)
int stringdiffs::FindNextMatchInWords2(const word & needword1, int bw2) const
{
  int cc1 = m_words2.GetSize();
	while (bw2 < cc1)
	{
		if (AreWordsSame(needword1, m_words2[bw2]))
			return bw2;
		++bw2;
	}
	return -1;
}

// @brief Find next word in m_words1 (starting at bw1) that matches needword2 (in m_words2)
int stringdiffs::FindNextMatchInWords1(const word & needword2, int bw1) const
{
  int cc1 = m_words1.GetSize();
	while (bw1 < cc1)
	{
		if (AreWordsSame(m_words1[bw1], needword2))
			return bw1;
		++bw1;
	}
	return -1;
}

/**
* @brief Add all different elements between lines to the m_pDiffs list.
*/
void stringdiffs::BuildWordDiffList()
{
	BuildWordsArray(m_str1, &m_words1);
	BuildWordsArray(m_str2, &m_words2);
  int len1 = m_str1.GetLength();
  int len2 = m_str2.GetLength();

	//INT_PTR nWordsCount = min(m_words1.GetSize(), m_words2.GetSize());
  INT_PTR wc1 = m_words1.GetSize();
  INT_PTR wc2 = m_words2.GetSize();
	INT_PTR nDiffStartAtWord1 = -1;
  INT_PTR nDiffStartAtWord2 = -1;
	bool bOutWordDiff = true;

	// Compare each word against its matching word on the second list. Each
	// difference starts from the first non-matching word and ends in the
	// predecessor of the first matching word.
  INT_PTR w1 = 0;
  INT_PTR w2 = 0;
  //INT_PTR nWord;
	//for (nWord = 0; nWord < nWordsCount; ++nWord)
  for (; (w1 < wc1) && (w2 < wc2); ++w1, ++w2)
	{
		if (bOutWordDiff)
		{
			if (!AreWordsSame(m_words1[w1], m_words2[w2]))
			{
				nDiffStartAtWord1 = w1;
        nDiffStartAtWord2 = w2;
				bOutWordDiff = findSync(w1, w2);
        if (bOutWordDiff)
        {
          int s1 = m_words1[nDiffStartAtWord1].start; //len1;
          int e1 = s1-1;
          int s2 = m_words2[nDiffStartAtWord2].start; //len2;
          int e2 = s2-1;
          if (w1 > nDiffStartAtWord1)
            e1 = m_words1[w1-1].end;
          if (w2 > nDiffStartAtWord2)
            e2 = m_words2[w2-1].end;
          AddDiff(s1, e1, s2, e2);
        }
			}
		}
		else
		{
			/*if (AreWordsSame(m_words1[nWord], m_words2[nWord]))
			{
				AddWordDiff(nDiffStartAtWord, nWord - 1);
				bInWordDiff = false;
			}*/
      bOutWordDiff = AreWordsSame(m_words1[w1], m_words2[w2]);
      if (bOutWordDiff)
        AddDiff(m_words1[nDiffStartAtWord1].start, m_words1[w1-1].end, m_words2[nDiffStartAtWord2].start, m_words2[w2-1].end);
      else
      {
        bOutWordDiff = findSync(w1, w2);
        if (bOutWordDiff)
          AddDiff(m_words1[nDiffStartAtWord1].start, m_words1[w1-1].end, m_words2[nDiffStartAtWord2].start, m_words2[w2-1].end);
      }
		}
	}

	// Ops, we left an open word difference, better close it. This could
	// happen when there is only one difference. Either a line with a single
	// word or reaching end of line without finding matching word.
	if (bOutWordDiff)
  {
    if (w1 < wc1)
      AddDiff(m_words1[w1].start, len1-1, len2, len2-1);
    if (w2 < wc2)
      AddDiff(len1, len1-1, m_words2[w2].start, len2-1);
  }
  else
		AddDiff(m_words1[nDiffStartAtWord1].start, m_words1[wc1-1].end, m_words2[nDiffStartAtWord2].start, m_words2[wc2-1].end);

	//HandleLeftOvers(nWordsCount, bInWordDiff);
}

/**
* @brief Add a difference using offsets.
*
* @param s1 [in] Left difference start offset.
* @param e1 [in] Left difference end offset.
* @param s2 [in] Right difference start offset.
* @param e2 [in] Right difference end offset.
*/
void stringdiffs::AddDiff(int s1, int e1, int s2, int e2)
{
	wdiff wdf(s1, e1, s2, e2);
	m_pDiffs->Add(wdf);
}

/**
* @brief Add a word to a words array.
*
* @param str [in] A string which contains the word to add.
* @param words [in] A words arrays.
* @param s [in] Start of word offset.
* @param e [in] End of word offset.
* @param ws [in] True if the word is a whitespace word.
*/
void stringdiffs::AddWord(const CString & str, wordarray * words, int s, int e, bool ws) const
{
	if (!ws || (m_whitespace == WHITESPACE_COMPARE_ALL))
	{
		word wd(s, e, hash(str, s, e));
		words->Add(wd);
	}
}

/**
* @brief Add a difference using words range.
*
* @param start [in] Start difference at word index.
* @param end [in] End difference at word index.
*/
/*void stringdiffs::AddWordDiff(int start, int end)
{
	AddDiff(m_words1[start].start, m_words1[end].end,
		m_words2[start].start, m_words2[end].end);
}*/

/**
* @brief Extend the last difference to a new end.
*
* @param bLeftSide [in] true if to extend left side, false to right side.
* @param nEnd [in] New end offset.
*
* @return true if difference was extended, false if no difference exist.
*/
/*bool stringdiffs::ExtendLastDiff(bool bLeftSide, int nEnd)
{
	const INT_PTR nLastDiff = m_pDiffs->GetUpperBound();
  bool res = nLastDiff >= 0;
	if (res)
	{
		wdiff& wdf = m_pDiffs->GetAt(nLastDiff);
		wdf.end[bLeftSide] = nEnd;
	}

	return res;
}*/

/**
* @brief Handle the case where word lists are not in the same size.
*
* This method highlight the remaining words, in other words, each word that
* don't have a counter word. All the extra words are mark as a single
* difference. If the last compared words are not equal, the last difference
* is extend to include all additional words.
*
* @param nLastWord [in] Word index where processing has stopped.
* @param bInWordDiff [in] true if last compared words are different.
*/
/*void stringdiffs::HandleLeftOvers(int nLastWord, bool bInWordDiff)
{
	if (nLastWord < m_words1.GetSize())
	{
		// Left side has more words.
		const int nLeftEnd = m_words1[m_words1.GetUpperBound()].end;

		if (!bInWordDiff || !ExtendLastDiff(false, nLeftEnd))
		{
			AddDiff(m_words1[nLastWord].start,
				nLeftEnd,
				m_str2.GetLength(),	-1);
		}
	}
	else if (nLastWord < m_words2.GetSize())
	{
		// Right side has more words.
		const int nRightEnd = m_words2[m_words2.GetUpperBound()].end;

		if (!bInWordDiff || !ExtendLastDiff(true, nRightEnd))
		{
			AddDiff(m_str1.GetLength(), -1,
				m_words2[nLastWord].start,
				nRightEnd);
		}
	}
}*/

/**
 * @brief Break line into constituent words
 */
void stringdiffs::BuildWordsArray(const CString & str, wordarray * words) const
{
	int begin = -1;
  int cc = str.GetLength();
	// state when we are looking for next word
  for (int i = 0; i < cc; ++i)
  {
    bool bbr = isWordBreak(m_breakType, str[i]);
    bool aspace = isSafeWhitespace(str[i]);
    if (begin < 0)
    {
      if (!aspace)
      {
        if (bbr)
          AddWord(str, words, i);
        else
          begin = i;
      }
    }
    else if (aspace || bbr)
    {
      AddWord(str, words, begin, i);
      if (bbr)
        AddWord(str, words, i);
      begin = -1;
    }
  }
  if ((begin >= 0) && (begin < cc))
    AddWord(str, words, begin, cc);
}

// diffutils hash

/* Rotate a value n bits to the left. */
#define UINT_BIT ((UINT)(sizeof (UINT) * CHAR_BIT))
#define ROL(v, n) ((v) << (n) | (v) >> (UINT_BIT - (n)))
/* Given a hash value and a new character, return a new hash value. */
#define HASH(h, c) ((c) + ROL (h, 7u))

UINT stringdiffs::hash(const CString & str, int begin, int end) const
{
	UINT h = 0;
	for (int i=begin; i<=end; ++i)
	{
		UINT ch = (UINT)str[i];
    h += HASH(h, (m_case_sensitive) ? ch : (UINT)_totupper(ch));
	}
	return h;
}

/**
 * @brief Compare two words (by reference to original strings)
 */
bool
stringdiffs::AreWordsSame(const word & word1, const word & word2) const
{
	if (word1.hash != word2.hash)
		return false;
  int c = word1.length();
	if (c != word2.length())
		return false;
  if (m_case_sensitive)
    return !_tcsncmp(&((LPCTSTR)m_str1)[word1.start], &((LPCTSTR)m_str2)[word2.start], c);
  return !_tcsnicmp(&((LPCTSTR)m_str1)[word1.start], &((LPCTSTR)m_str2)[word2.start], c);
}

/**
 * @brief Return true if chars match
 *
 * Caller must not call this for lead bytes
 */
bool matchchar(TCHAR ch1, TCHAR ch2, bool casitive)
{
	if (casitive)
		return ch1==ch2;
	return _totupper(ch1)==_totupper(ch2);
}

/**
 * @brief Return true if characters match
 */
bool
stringdiffs::caseMatch(TCHAR ch1, TCHAR ch2) const
{
	return matchchar(ch1, ch2, m_case_sensitive);
}

/**
 * @brief Return pointer to last character of specified string (handle MBCS)
 *
 * If the last byte is a broken multibyte (ie, a solo lead byte), this returns previous char
 */
LPCTSTR LastChar(LPCTSTR psz, int len)
{
	if (len <= 0) return psz;
  LPCTSTR lastValid = psz + (len - 1);
#ifdef _UNICODE
  return lastValid;
#else
	if (!_getmbcp()) return lastValid;

	LPCTSTR prev=psz;
	while (psz<lastValid)
	{
		prev = psz;
		psz = CharNext(psz);
	}
	if (psz==lastValid && !IsLeadByte(*psz))
		return psz;
	return prev; // last character was multibyte or broken multibyte
#endif
}

/**
 * @brief advance current pointer over whitespace, until not whitespace or beyond end
 * @param pcurrent [in,out] current location (to be advanced)
 * @param end [in] last valid position (only go one beyond this)
 */
void AdvanceOverWhitespace(LPCTSTR &pcurrent, LPCTSTR end)
{
	// advance over whitespace
	while (pcurrent <= end && isSafeWhitespace(*pcurrent))
		++pcurrent; // DBCS safe because of isSafeWhitespace above
}

/**
 * @brief back current pointer over whitespace, until not whitespace or at start
 * @param pcurrent [in,out] current location (to be backed up)
 * @param start [in] first valid position (do not go before this)
 *
 * NB: Unlike AdvanceOverWhitespace, this will not go over the start
 * This because WinMerge doesn't need to, and also CharPrev cannot easily do so
 */
void RetreatOverWhitespace(LPCTSTR &pcurrent, LPCTSTR start)
{
	// back over whitespace
	while (pcurrent > start && isSafeWhitespace(*pcurrent))
		pcurrent = CharPrev(start, pcurrent); // DBCS safe because of isSafeWhitespace above
}

/**
 * @brief Compute begin1,begin2,end1,end2 to display byte difference between strings str1 & str2
 * @param casitive [in] true for case-sensitive, false for case-insensitive
 * @param xwhite [in] This governs whether we handle whitespace specially (see WHITESPACE_COMPARE_ALL, WHITESPACE_IGNORE_CHANGE, WHITESPACE_IGNORE_ALL)
 *
 * Assumes whitespace is never leadbyte or trailbyte!
 */
void
sd_ComputeByteDiff(CString & str1, CString & str2, 
		   bool casitive, int xwhite, 
		   int &begin1, int &begin2, int &end1, int &end2)
{
	// Set to sane values
	// Also this way can distinguish if we set begin1 to -1 for no diff in line
	begin1 = end1 = begin2 = end2 = 0;

	int len1 = str1.GetLength();
	int len2 = str2.GetLength();

	LPCTSTR pbeg1 = (LPCTSTR)str1;
	LPCTSTR pbeg2 = (LPCTSTR)str2;

	if (len1 == 0 || len2 == 0)
	{
		if (len1 == len2)
			begin2 = begin1 = -1;
		end1 = len1 - 1;
		end2 = len2 - 1;
		return;
	}

	// cursors from front, which we advance to beginning of difference
	LPCTSTR py1 = pbeg1;
	LPCTSTR py2 = pbeg2;

	// pen1,pen2 point to the last valid character (broken multibyte lead chars don't count)
	LPCTSTR pen1 = LastChar(py1, len1);
	LPCTSTR pen2 = LastChar(py2, len2);

	if (xwhite != WHITESPACE_COMPARE_ALL)
	{
		// Ignore leading and trailing whitespace
		// by advancing py1 and py2
		// and retreating pen1 and pen2
		while (py1 < pen1 && isSafeWhitespace(*py1))
			++py1; // DBCS safe because of isSafeWhitespace above
		while (py2 < pen2 && isSafeWhitespace(*py2))
			++py2; // DBCS safe because of isSafeWhitespace above
    if ((pen1 == (pbeg1 + len1 - 1)) && (pen2 == (pbeg2 + len2 - 1))
			|| len1 && len2 && (pbeg1[len1] == pbeg2[len2]))
		/*if ((pen1 < pbeg1 + len1 - 1 || pen2 < pbeg2 + len2 -1)
			&& (!len1 || !len2 || pbeg1[len1] != pbeg2[len2]))
		{
			// mismatched broken multibyte ends
		}
		else*/
		{
			while (pen1 > py1 && isSafeWhitespace(*pen1))
				pen1 = CharPrev(py1, pen1);
			while (pen2 > py2 && isSafeWhitespace(*pen2))
				pen2 = CharPrev(py2, pen2);
		}
	}

	bool alldone = false;
	// Advance over matching beginnings of lines
	// Advance py1 & py2 from beginning until find difference or end
	while (true)
	{
		// Potential difference extends from py1 to pen1 and py2 to pen2

		// Check if either side finished
    alldone = (py1 > pen1);
    bool bb2 = (py2 > pen2);
		if (alldone && bb2)
		{
			begin1 = end1 = begin2 = end2 = -1;
			//alldone = true;
			break;
		}
    alldone |= bb2;
		if (alldone)
			break;

		// handle all the whitespace logic (due to WinMerge whitespace settings)
		if (xwhite!=WHITESPACE_COMPARE_ALL && isSafeWhitespace(*py1))
		{
			if (xwhite==WHITESPACE_IGNORE_CHANGE && !isSafeWhitespace(*py2))
			{
				// py1 is white but py2 is not
				// in WHITESPACE_IGNORE_CHANGE mode,
				// this doesn't qualify as skippable whitespace
				break; // done with forward search
			}
			// gobble up all whitespace in current area
			AdvanceOverWhitespace(py1, pen1); // will go beyond end
			AdvanceOverWhitespace(py2, pen2); // will go beyond end
			continue;
		}
		if (xwhite!=WHITESPACE_COMPARE_ALL && isSafeWhitespace(*py2))
		{
			if (xwhite==WHITESPACE_IGNORE_CHANGE && !isSafeWhitespace(*py1))
			{
				// py2 is white but py1 is not
				// in WHITESPACE_IGNORE_CHANGE mode,
				// this doesn't qualify as skippable whitespace
				break; // done with forward search
			}
			// gobble up all whitespace in current area
			AdvanceOverWhitespace(py1, pen1); // will go beyond end
			AdvanceOverWhitespace(py2, pen2); // will go beyond end
			continue;
		}

		// Now do real character match
		if (IsLeadByte(*py1))
		{
			if (!IsLeadByte(*py2))
				break; // done with forward search
			// DBCS (we assume if a lead byte, then character is 2-byte)
			if (!(py1[0] == py2[0] && py1[1] == py2[1]))
				break; // done with forward search
			py1 += 2; // DBCS specific
			py2 += 2; // DBCS specific
		}
		else
		{
			if (IsLeadByte(*py2))
				break; // done with forward search
			if (!matchchar(*py1, *py2, casitive))
				break; // done with forward search
			++py1; // DBCS safe b/c we checked above
			++py2; // DBCS safe b/c we checked above
		}
	}

	// Potential difference extends from py1 to pen1 and py2 to pen2

	// Store results of advance into return variables (begin1 & begin2)
	// -1 in a begin variable means no visible diff area
	begin1 = (py1 > pen1) ? -1 : (py1 - pbeg1);
	begin2 = (py2 > pen2) ? -1 : (py2 - pbeg2);

	if (alldone)
	{
		end1 = pen1 - pbeg1;
		end2 = pen2 - pbeg2;
	}
	else
	{
		LPCTSTR pz1 = pen1;
		LPCTSTR pz2 = pen2;

		// Retreat over matching ends of lines
		// Retreat pz1 & pz2 from end until find difference or beginning
		while (true)
		{
			// Potential difference extends from py1 to pz1 and from py2 to pz2

			// Check if either side finished
      bool bb1 = (pz1 < py1);
      bool bb2 = (pz2 < py2);
			if (bb1 && bb2)
			{
				begin1 = end1 = begin2 = end2 = -1;
				break;
			}
			if (bb1 || bb2)
				break;

			// handle all the whitespace logic (due to WinMerge whitespace settings)
			if (xwhite!=WHITESPACE_COMPARE_ALL && isSafeWhitespace(*pz1))
			{
				if (xwhite==WHITESPACE_IGNORE_CHANGE && !isSafeWhitespace(*pz2))
				{
					// pz1 is white but pz2 is not
					// in WHITESPACE_IGNORE_CHANGE mode,
					// this doesn't qualify as skippable whitespace
					break; // done with reverse search
				}
				// gobble up all whitespace in current area
				RetreatOverWhitespace(pz1, py1); // will not go over beginning
				RetreatOverWhitespace(pz2, py2); // will not go over beginning
				continue;

			}
			if (xwhite!=WHITESPACE_COMPARE_ALL && isSafeWhitespace(*pz2))
			{
				if (xwhite==WHITESPACE_IGNORE_CHANGE && !isSafeWhitespace(*pz1))
				{
					// pz2 is white but pz1 is not
					// in WHITESPACE_IGNORE_CHANGE mode,
					// this doesn't qualify as skippable whitespace
					break; // done with reverse search
				}
				// gobble up all whitespace in current area
				RetreatOverWhitespace(pz1, py1); // will not go over beginning
				RetreatOverWhitespace(pz2, py2); // will not go over beginning
				continue;
			}

			// Now do real character match
			if (IsLeadByte(*pz1))
			{
				if (!IsLeadByte(*pz2))
					break; // done with forward search
				// DBCS (we assume if a lead byte, then character is 2-byte)
				if (!(pz1[0] == pz2[0] && pz1[1] == pz2[1]))
					break; // done with forward search
			}
			else
			{
				if (IsLeadByte(*pz2))
					break; // done with forward search
				if (!matchchar(*pz1, *pz2, casitive))
					break; // done with forward search
			}
			// decrement pz1 and pz2
			if (pz1 == pbeg1)
				pz1 = pbeg1 - 1; // earlier than pbeg1 signifies no difference
			else
				pz1 = CharPrev(pbeg1, pz1);
			if (pz2 == pbeg2)
				pz2 = pbeg2 - 1; // earlier than pbeg1 signifies no difference
			else
				pz2 = CharPrev(pbeg2, pz2);
		}

		// Store results of advance into return variables (end1 & end2)
		if (pz1 < pbeg1)
			begin1 = -1; // no visible diff in line 1
		else
			end1 = pz1 - pbeg1;
		if (pz2 < pbeg2)
			begin2 = -1; // no visible diff in line 2
		else
			end2 = pz2 - pbeg2;
    if (end1 < begin1)
      begin1 = -1;
		if (end2 < begin2)
      begin2 = -1;
	}
}
